/*
 * Convert from a Z80 to a 68K file - conversion code.
 *
 * Copyright (C) 2005 Nick Tarleton.
 * Licensed under the BSD license; see BSD.txt.
 */

import java.io.*;
import java.util.Arrays;

final class Converter {
	/* Shared parts (pass data from input to output) */
	// type of input file, one of the TYPE_ constants
	private static String fileName;
	private static String filePath;
	private static int inType;
	private static String varName;
	private static byte[] varData;

	static void doConvert(File file) throws IOException
	{
		fileName = file.getName();
		filePath = file.getPath();
		Main.writeMessage("Converting: " + filePath + '\n');

		readInput(file);
		writeOutput();
	}

	/* Input-specific parts */
	private static RandomAccessFile inFile;

	private static byte[] readBytes(int n) throws IOException
	{
		byte[] result = new byte[n];
		inFile.readFully(result);
		return result;
	}

	private static short readShortLE() throws IOException
	{
		short rv = (short) inFile.readUnsignedByte();
		rv |= inFile.readUnsignedByte() << 8;
		return rv;
	}

	// magic constants
	// TYPE_x has no meaning outside MulTI
	// HEADER_x is the expected header from the input file
	private static final int TYPE_83PLUS = 0x84;
	private static final byte[] HEADER_83X = { 0x1A, 0x0A, 0x00 };

	static void readInput(File file) throws IOException
	{
		byte[] data;

		byte[] expectedHeader;
		String ext = fileName.substring(fileName.lastIndexOf('.') + 1);
		if(ext.equalsIgnoreCase("8xp")) {
			Main.writeMessage("File type from extension: TI-83+\n");
			inType = TYPE_83PLUS;
			expectedHeader = HEADER_83X;
		} else {
			throw new RuntimeException("Unknown extension: " + ext);
		}

		inFile = new RandomAccessFile(file, "r");

		inFile.seek(0x08); // seek to filetype header
		data = readBytes(3);
		if(!Arrays.equals(data, expectedHeader))
			throw new RuntimeException("Type does not match extension");

		inFile.seek(0x37); // seek to var header length
		int varHeaderLength = (int) readShortLE();

		inFile.seek(0x3c); // seek to program name
		StringBuffer sb = new StringBuffer();
		byte b;
		while(((b = inFile.readByte()) != 0) && (sb.length() < 8))
			sb.append((char) b);
		varName = sb.toString();
		Main.writeMessage("Variable name: " + varName + '\n');

		inFile.seek(0x39 + varHeaderLength + 2); // seek to prog length
		int progLength = (int) readShortLE();
		Main.writeMessage("Program length: " + progLength + '\n');

		varData = readBytes(progLength);

		// -0x45 = 0xBB
		if(varData[0] != -0x45 || varData[1] != 0x6D)
			throw new RuntimeException("Not an ASM program");
	}

	/* Output-specific parts */
	private static RandomAccessFile outFile;

	static void writePaddedString(String str, int padLen) throws IOException
	{
		if(str.length() > padLen)
			throw new IllegalArgumentException('\"' + str
				+ "\" longer than padLen of " + padLen);
		outFile.writeBytes(str);
		byte[] padding = new byte[padLen - str.length()];
		Arrays.fill(padding, (byte) 0);
		outFile.write(padding);
	}

	private static void writeShortLE(short s) throws IOException
	{
		outFile.writeByte((int) (s & 0xff));
		outFile.writeByte((int) (s >> 8));
	}

	private static void writeIntLE(int s) throws IOException
	{
		outFile.writeByte(s & 0xff);
		outFile.writeByte((s >> 8) & 0xff);
		outFile.writeByte((s >> 16) & 0xff);
		outFile.writeByte(s >> 24);
	}

	// FIXME will need major patching to support 92/V200
	static void writeOutput() throws IOException
	{
		StringBuffer sb = new StringBuffer();
		sb.append(filePath.substring(0,
			filePath.lastIndexOf(File.separatorChar) + 1));
		sb.append(varName);
		sb.append(".89y");
		String outFileName = sb.toString().toLowerCase();
		Main.writeMessage("Output file: " + outFileName + '\n');

		outFile = new RandomAccessFile(outFileName, "rw");
		outFile.setLength(0);

		// header
		outFile.writeBytes("**TI89**"); // tag
		writeShortLE((short) 1); // maybe number of folders?
		writePaddedString("main", 8); // folder
		writePaddedString("Converted program for MulTI", 40); // comment
		writeShortLE((short) 1); // number of vars
		writeIntLE(0x52);
			// offset of data (actually 32 bits)
		writePaddedString(varName, 8);
		writeIntLE(0x1c); // file type (other)
		writeIntLE(0); // placeholder for file size
		writeShortLE((short) 0x5AA5); // maybe start of data tag?

		// data
		byte[] outData = new byte[10 + varData.length + 7];
		outData[0] = (byte) 0; // filler
		outData[1] = (byte) 0;
		outData[2] = (byte) 0;
		outData[3] = (byte) 0;
		// length
		outData[4] = (byte) ((outData.length - 6) >> 8);
		outData[5] = (byte) ((outData.length - 6) & 0xff);
		// begin actual file contents
		outData[6] = (byte) inType; // target calc type
		outData[7] = (byte) 0; // alignment
		outData[8] = (byte) (varData.length >> 8);
		outData[9] = (byte) (varData.length & 0xff);
		System.arraycopy(varData, 0, outData, 10, varData.length);
		// end actual file contents
		outData[outData.length - 7] = (byte) 0; // extension
		outData[outData.length - 6] = (byte) 'm';
		outData[outData.length - 5] = (byte) 'l';
		outData[outData.length - 4] = (byte) 'T';
		outData[outData.length - 3] = (byte) 'I';
		outData[outData.length - 2] = (byte) 0;
		outData[outData.length - 1] = (byte) 0xf8;
		outFile.write(outData);

		// checksum = sum of all data bytes
		int checksum = 0;
		for(int i = 0; i < outData.length; i++)
			checksum += ((short) outData[i]) & 0xff;
		writeShortLE((short) (checksum & 0xffff));

		// finally write file size
		outFile.seek(0x4c);
		writeIntLE((int) outFile.length());
		
		outFile.close();
		Main.writeMessage("Done.\n");
	}

	private Converter() { /* don't instantiate me */ }
}
