/*
 * Convert from a Z80 to a 68K file - main program.
 *
 * Copyright (C) 2005 Nick Tarleton.
 * Licensed under the BSD license; see BSD.txt.
 */

import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.plaf.metal.MetalIconFactory;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.File;

final class TIFileFilter extends FileFilter {
	public String getDescription()
	{
		return "TI-xx files";
	}

	public boolean accept(File f)
	{
		if(f.isDirectory())
			return true;

		String name = f.getName();
		int dotIdx = name.lastIndexOf('.');
		if(dotIdx == -1)
			return false;

		String ext = name.substring(dotIdx + 1);
		if(ext.equalsIgnoreCase("8xp"))
			return true;

		return false;
	}
}

final class ConvertButtonListener implements ActionListener {
	public void actionPerformed(ActionEvent event)
	{
		JFileChooser chooser = new JFileChooser();
		chooser.setFileFilter(new TIFileFilter());
		chooser.setMultiSelectionEnabled(true);
		int rv = chooser.showOpenDialog(Main.mainWin);
		if(rv == JFileChooser.APPROVE_OPTION) {
			Main.clearMessages();
			File[] files = chooser.getSelectedFiles();
			int success = 0, failure = 0;
			for(int i = 0; i < files.length; i++) {
				try {
					Converter.doConvert(files[i]);
					success++;
				} catch(Exception e) {
					Main.writeMessage("ERROR: "
						+ e.getMessage() + '\n');
					failure++;
				}
				Main.writeMessage("----------------------------------------\n");
			}
			Main.writeMessage("Done: " + files.length + " files, "
				+ success + " successes, "
				+ failure + " failures.");
		}
	}
}

public final class Main {
	static JFrame mainWin;
	private static JButton doIt;
	private static JTextArea textArea;

	private static boolean runningOnConsole;

	static void clearMessages()
	{
		textArea.setText("");
	}

	static void writeMessage(String message)
	{
		if(runningOnConsole)
			System.out.print(message);
		else
			textArea.append(message);
	}

	public static void main(String[] args)
	{
		if(args.length > 0) {
			runningOnConsole = true;
			int success = 0, failure = 0;
			for(int i = 0; i < args.length; i++) {
				try {
					Converter.doConvert(new File(args[i]));
					success++;
				} catch(Exception e) {
					writeMessage("ERROR: "
						+ e.getMessage() + '\n');
					failure++;
				}
				writeMessage("----------------------------------------\n");
			}
			writeMessage("Done: " + args.length + " files, "
				+ success + " successes, "
				+ failure + " failures.\n");
			return;
		} else {
			runningOnConsole = false;
			System.out.println(
"You can also use this program from the command line, for example:\njava -jar convert.jar file1.8xp file2.8xp"
			);
		}

		// components
		mainWin = new JFrame("MulTI File Converter");
		doIt = new JButton("Convert file(s)...",
			new MetalIconFactory.FolderIcon16());
		doIt.addActionListener(new ConvertButtonListener());
		textArea = new JTextArea(30, 50);
		textArea.setEditable(false);

		// layout
		Box hbox = new Box(BoxLayout.X_AXIS);
		hbox.add(doIt);
		mainWin.getContentPane().setLayout(new BoxLayout(mainWin.getContentPane(),
			BoxLayout.Y_AXIS));
		mainWin.getContentPane().add(hbox);
		mainWin.getContentPane().add(Box.createVerticalStrut(5));
		mainWin.getContentPane().add(new JSeparator(SwingConstants.HORIZONTAL));
		mainWin.getContentPane().add(Box.createVerticalStrut(5));
		mainWin.getContentPane().add(textArea);

		// finish
		mainWin.pack();
		mainWin.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		writeMessage("Ready.\n");
		mainWin.setVisible(true);
	}

	private Main() { /* don't instantiate me */ }
}
