/*
 * 68K generation stuff.
 *
 * Copyright (C) 2005 Nick Tarleton.
 * Licensed under the GNU General Public License; see COPYING.txt.
 */

#ifndef _multi_68K_GEN_H
#define _multi_68K_GEN_H

/* EA MODES */

#define SIZE_B 0
#define SIZE_W 1
#define SIZE_L 2

#define MODE_Dn 0
#define MODE_An 1
#define MODE_pAn 2
#define MODE_pAn_inc 3
#define MODE_pAn_dec 4
#define MODE_pAn_disp16 5
#define MODE_pAn_disp8_Rn 6
#define MODE_OTHER 7

#define MODE_OTHER_pimmW 0
#define MODE_OTHER_pimmL 1
// XXX the following two modes can only be used with a few instructions
#define MODE_OTHER_pPC_disp16 2
#define MODE_OTHER_pPC_disp8_Dn 3
#define MODE_OTHER_imm 4
// IMPORTANT NOTE: The 68K design requires that all instructions be
// word-aligned. Therefore, when using a byte-sized immediate, the value has
// to be emitted as a word.

#define RDATA 0
#define RADDR 1
#define RWORD 0
#define RLONG 1

#define emit_pAn_disp8_Rn(disp, r_type, r_length, r_num) \
	emit_w(((r_type) << 15) | ((r_num) << 12) | ((r_length) << 11) | (disp))
#define emit_pPC_disp8_Dn(disp, reg) \
	emit_w((reg << 12) | disp)

/* INSTRUCTIONS */

/* Move */

#define MOVE_SIZE_B 1
#define MOVE_SIZE_W 3
#define MOVE_SIZE_L 2

#define emit_MOVE(size, src_mode, src_reg, dest_mode, dest_reg) \
	emit_w(((size) << 12) | ((dest_reg) << 9) | ((dest_mode) << 6) | ((src_mode) << 3) | (src_reg))

/* Binary Operations */
#define emit_bin_op_ea_Dn(size, op, src_mode, src_reg, dest_reg) \
	emit_w(((op) << 8) | ((dest_reg) << 9) | ((size) << 6) | ((src_mode) << 3) | (src_reg))
#define emit_bin_op_imm_ea(size, op, dest_mode, dest_reg) \
	emit_w(((op) << 8) | ((size) << 6) | ((dest_mode) << 3) | (dest_reg))
#define emit_bin_op_ext_Dn_Dn(size, op, src_reg, dest_reg) \
	emit_w(((op) << 12) | ((dest_reg) << 9) | (1 << 8) | ((size) << 6) | (src_reg))
#define emit_bin_op_imm_ea_Q(size, op, data, dest_mode, dest_reg) \
	emit_w(0x5000 | ((data) << 9) | ((op) << 8) | ((size) << 6) | ((dest_mode) << 3) | (dest_reg))

#define ADD 0xD0
#define ADD_reverse 0xD1
#define ADDI 6
#define ADDX 13
#define ADDQ 0
#define AND 0xC0
#define AND_reverse 0xC1
#define ANDI 2
#define CMP 0xB0
#define CMPI 12
#define EOR 0xB1 // XXX this is in reverse order (Dn,ea)
#define EORI 10
#define SUB 0x90
#define SUB_reverse 0x91
#define SUBI 4
#define SUBX 9
#define SUBQ 1
#define OR 0x80
#define OR_reverse 0x81
#define ORI 0

/* Shift/Rotation */

#define RIGHT 0
#define LEFT 1

// Notes: count of 8 is specified as 0
#define emit_RoSh_imm_Dn(op, dir, size, count, reg) \
	emit_w(op | ((dir) << 8) | ((size) << 6) | ((count) << 9) | (reg))
#define emit_ROd_imm_Dn(dir, size, count, reg) \
	emit_RoSh_imm_Dn(ROd, dir, size, count, reg)
#define emit_ROXd_imm_Dn(dir, size, count, reg) \
	emit_RoSh_imm_Dn(ROXd, dir, size, count, reg)
#define emit_ASd_imm_Dn(dir, size, count, reg) \
	emit_RoSh_imm_Dn(ASd, dir, size, count, reg)
#define emit_LSd_imm_Dn(dir, size, count, reg) \
	emit_RoSh_imm_Dn(LSd, dir, size, count, reg)

#define ROd 0xE018
#define ROXd 0xE010
#define ASd 0xE000
#define LSd 0xE008

// Difference between arithmetic and logical shifts:
// Left shift is the same: low bits are filled with zeroes.
// Right shift: high bits are filled with zeroes in logical, MSb in arithmetic

/* Bit */
// XXX: long size with Dn, byte size with all others
#define emit_bitop_imm(op, bitnr, mode, reg) \
	do { emit_w((op) | ((mode) << 3) | (reg)); emit_w(bitnr); } while(0)
#define BCLR 0x0880
#define BSET 0x08C0
#define BTST 0x0800

/* Flow */

// displacement is relative to the word following the branch
// (the next instruction, or the word-size displacement)
// use disp of 0 to indicate word displacement follows
#define emit_BRA(disp) \
	emit_w(0x6000 | disp)
#define emit_BSR(disp) \
	emit_w(0x6100 | disp)
#define emit_Bcc(condition, disp) \
	emit_w(0x6000 | ((condition) << 8) | (disp))
// always used with word disp
#define emit_DBcc(condition, reg, disp) \
	do { emit_w(0x50C8 | ((condition) << 8) | (reg)); emit_w(disp); } while(0)
#define emit_DBRA(reg, disp) emit_DBcc(CC_F, reg, disp)

// for generating negative displacements
#define NEG_BYTE(disp) ((BYTE)(-((BYTE)(disp))))
#define NEG_WORD(disp) ((WORD)(-((WORD)(disp))))

// condition codes
#define CC_T 0 // always true
#define CC_F 1 // always false
#define CC_HI 2 // high
#define CC_LS 3 // low or same
#define CC_CC 4 // carry clear
#define CC_CS 5 // carry set
#define CC_NE 6 // not equal (zero clear)
#define CC_EQ 7 // equal (zero set)
#define CC_VC 8 // overflow clear
#define CC_VS 9 // overflow set
#define CC_PL 10 // plus
#define CC_MI 11 // minus
#define CC_GE 12 // >=
#define CC_LT 13 // <
#define CC_GT 14 // >
#define CC_LE 15 // <=

#define emit_JMP(mode, reg) \
	emit_w(0x4EC0 | ((mode) << 3) | (reg))
#define emit_JSR(mode, reg) \
	emit_w(0x4E80 | ((mode) << 3) | (reg))

// convenience
#define emit_JSR_longimm(where) \
	do { emit_JSR(MODE_OTHER, MODE_OTHER_pimmL); emit_l((DWORD) where); } while(0)
#define emit_JMP_longimm(where) \
	do { emit_JMP(MODE_OTHER, MODE_OTHER_pimmL); emit_l((DWORD) where); } while(0)

#define emit_RTS() emit_w(0x4E75)

/* CCR */
#define emit_MOVE_from_SR(dest_mode, dest_reg) \
	emit_w(0x40C0 | ((dest_mode) << 3) | dest_reg)
#define emit_MOVE_to_CCR(src_mode, src_reg) \
	emit_w(0x44C0 | ((src_mode) << 3) | src_reg)
#define emit_EORI_to_CCR(value) \
	emit_l(0x0A3C0000 | value)
#define emit_ORI_to_CCR(value) \
	emit_l(0x003C0000 | value)

/* Misc */

#define emit_NOP() emit_w(0x4E71)

#define emit_NOT(size, mode, reg) \
	emit_w(0x4600 | ((size) << 6) | ((mode) << 3) | (reg))

#define emit_NEG(size, mode, reg) \
	emit_w(0x4400 | ((size) << 6) | ((mode) << 3) | (reg))

#define emit_CLR(size, mode, reg) \
	emit_w(0x4200 | ((size) << 6) | ((mode) << 3) | (reg))

#define emit_EXG(mode, reg1, reg2) \
	emit_w(0xC100 | ((reg1) << 9) | ((mode) << 3) | (reg2))
#define Dn_Dn 8
#define An_An 9
#define Dn_An 17

#endif
