/*
 * Main program file; provides UI and links everything together.
 *
 * Copyright (C) 2006 Nick Tarleton.
 * Licensed under the GNU General Public License; see COPYING.txt.
 */

#define MULTI_VERSION_STR "0.1\x80" // \x80 = alpha

#define USE_TI89
#define SAVE_SCREEN
#define ENABLE_ERROR_RETURN
#define NO_CALC_DETECT
#define NO_AMS_CHECK

#include <tigcclib.h>

#include "misc.h"
#include "target.h"
#include "xlate.h"

static void actual_main(void);
static void draw_initial_screen(void);

static const char* fname;
static INT_HANDLER save_int_1, save_int_5, save_int_6;
static char default_PRG_setting;
static char a5_switch;

extern void isr_5;
extern void isr_6;

void _main(void)
{
	// cleanup absolutely must be called, even if an error is thrown, so..
	TRY
		actual_main();
	ONERR
		// ER_EXIT is thrown by report_error()
		if(errCode != ER_EXIT) {
			cleanup();
			PASS;
		} else {
			// cleanup() is called in report_error
			return;
		}
	ENDTRY
	cleanup();
}

static void actual_main(void)
{
	SYM_STR fsymstr;
	HSym hsym;
	HANDLE fhandle;
	BYTE* fdata;
	WORD fsize;
	ESI argptr = top_estack;
	HARDWARE_PARM_BLOCK* hwinfo = (HARDWARE_PARM_BLOCK*)
		FL_getHardwareParmBlock();

	if(hwinfo->hardwareID == 9) {
		// if Titanium
		if(*(unsigned short*)((char*)_rom_call_addr(C0)+92)) {
			// if no HW3Patch
			report_error("HW3Patch required on a TI-89 Titanium. Get it at:\n\nhttp://ti89prog.kevinkofler.cjb.net/");
		}
		add_ghost_space = 0;
	} else {
		add_ghost_space = 0x40000;
	}

	if(hwinfo->len < 24 || hwinfo->gateArray == 1) // HW1
		default_PRG_setting = 0xB2;
	else
		default_PRG_setting = 0xCC;
	
	cur_target = NULL;
	save_int_1 = NULL;
	a5_switch = 0;
	xlate_zero_statics();

	/* get string arg */	
	if(RemainingArgCnt(argptr) != 1 || GetArgType(argptr) != STR_TAG)
		report_error("Usage: multi(\"filename\")");
	fname = GetStrnArg(argptr);
	argptr = top_estack;
	fsymstr = GetSymstrArg(argptr);

	/* dereference symbol */
	hsym = SymFind(fsymstr);
	if(hsym.folder == 0)
		report_error("Could not find file");
	fhandle = DerefSym(hsym)->handle;
	fdata = HLock(fhandle);
	if(fdata == NULL)
		report_error("Could not lock file");

	/* read size */
	fsize = *(WORD *)fdata;
	fdata += 2;

	/* check type */
	if((fdata[fsize - 1] != OTH_TAG) || strcmp(&fdata[fsize - 6], "mlTI")) {
		HeapUnlock(fhandle);
		report_error("Not a MulTI file");
	}
	
	/* check target calc */
	switch(fdata[0]) {
		case TAG_TI83PLUS:
			cur_target = &target_ti83plus;
			break;
		default:
			HeapUnlock(fhandle);
			report_error("Unknown target calc %x", fdata[0]);
	}

	/* go! */
	draw_initial_screen();
	cur_target->load(fdata + 2);
	HeapUnlock(fhandle);
	save_int_1 = GetIntVec(AUTO_INT_1);
	SetIntVec(AUTO_INT_1, DUMMY_HANDLER);
	save_int_5 = GetIntVec(AUTO_INT_5);
	SetIntVec(AUTO_INT_5, (INT_HANDLER) &isr_5);
	save_int_6 = GetIntVec(AUTO_INT_6);
	SetIntVec(AUTO_INT_6, (INT_HANDLER) &isr_6);
	pokeIO(0x600017, 252); // set auto-int 5 to 200Hz

	cur_target->run();
	pokeIO(0x600005, 0x10); // delay 1/4 sec to allow key release
	pokeIO(0x600005, 0x10);
	pokeIO(0x600005, 0x10);
	pokeIO(0x600005, 0x10);
	pokeIO(0x600005, 0x10);
}

CALLBACK void cleanup(void)
{
	pokeIO(0x600017, default_PRG_setting); // set auto-int 5 to default 20Hz
	free_remembered();
	if(save_int_1) {
		SetIntVec(AUTO_INT_6, save_int_6);
		SetIntVec(AUTO_INT_5, save_int_5);
		SetIntVec(AUTO_INT_1, save_int_1);
	}
	OSSetSR(0); // just in case
}

static void draw_str_centered(short y, const char* str)
{
	short x = (LCD_WIDTH - DrawStrWidth(str, FontGetSys())) / 2;
	DrawStr(x, y, str, A_NORMAL);
}

static void draw_initial_screen(void)
{
	static const WIN_RECT outside_viewport = {30, 16, 129, 83};
	char buf[64];

	ClrScr();

	// draw information
	FontSetSys(F_4x6);
	draw_str_centered(0, "MulTI version " MULTI_VERSION_STR);
	// \x10 = square
	draw_str_centered(7, "by Nick Tarleton \x10 nickptar@gmail.com");
	sprintf(buf, "Emulating: TI-83+ \x10 Program: %s", fname);
	draw_str_centered(LCD_HEIGHT - 7, buf);

	FontSetSys(F_6x8);
	draw_str_centered(42, "Initializing...");
}

/* AUTO_INT_5 routine to waste time */
asm("
isr_5:
	movem.l	%d0/%a0,-(%sp)
	lea	a5_switch(%pc),%a0
	not.b	(%a0)
	beq	1f
	move.b	#0x10,(0x600005)	| wait until next AI5 or AI6
1:	movem.l	(%sp)+,%d0/%a0
	rte
");

/* AUTO_INT_6 routine to abort when ON pressed
 * (thanks to the folks on the tigcc forums)
 */
asm("
isr_6:
	clr.b	(0x60001a)	| ack interrupt
	btst	#1,(0x60001a)	| was ON really pressed?
	beq	1f
	rte			| if not, return
1:	lea	0x4c00,%sp	| if so: reset supervisor SP
	move.w	#0,%sr		| unmask interrupts & go to user mode
	.word	0xA000+180	| throw ER_break
");
