/*
 * Translate Z80 instructions to 68K instructions.
 * This is the most important module.
 *
 * Copyright (C) 2006 Nick Tarleton.
 * Licensed under the GNU General Public License; see COPYING.txt.
 */

/* Here are some things to keep in mind, if you are brave enough to read this
 * code:
 **Numbers starting with '0' are octal. Octal is used because Z80 instructions
 * are made up of 3-bit fields.
 **In instructions with a displacement and an immediate value, the displacement
 * comes first.
 **Putting 0335 or 0375 before an instruction with H or L causes it to use the
 * high/low word of IX/IY. This is undocumented.
 **HOWEVER: In instructions with (HL), H or L separately, and a prefix,
 * IX/IY and a displacement are used for the address, but H/L are still used
 * as the byte registers.
 **0335 and 0375 are ignored before 0355.
 **Addresses on the Z80 are unsigned (duh). Displacements on the 68K are signed.
 * To compensate, A1 contains A0+0x8000.
 **However, displacements are signed on the Z80. Fortunately.
 **A 'block' in this program is not like a formal basic block; only
 * _unconditional_ jumps and returns end a block.
 **The 68K has two carry flags, C and X. Most instructions set them the same,
 * but some only affect C. X is used in the ADDX and SUBX instructions. This
 * gets annoying.
 *
 * Flags are handled in the following way:
 **ccr_saved is a flag in cur_state that starts at 0.
 **Any instruction that modifies the flags on the 68K but should not on the Z80
 * calls save_ccr(), which stores CCR to D7 and sets ccr_saved.
 **Any instruction that should modify the flags, including CALL and RST, calls
 * mod_ccr(), which clears ccr_saved.
 **Any instruction whose action depends on the flags calls restore_ccr(). If
 * ccr_saved is set, this loads CCR from D7. If ccr_saved is clear,
 * it does nothing.
 **PUSH AF stores A and CCR. Z80 code that tries to read/adjust the flags via
 * the stack will therefore not work.
 */

#include "ops.h"
#include "misc.h"
#include "xlate.h"
#include "target.h"
#include "glue.h"
#include "68k_gen.h"

/* Output */
#define CODE_EXPAND 16 // space to expand by each time expansion is necessary

static void expand(int x)
{
	if(cur_state->out_pos + x >= cur_state->alloc) {
		cur_state->alloc += CODE_EXPAND;
		cur_state->block->m68k_code =
			realloc_remember(cur_state->block->m68k_code,
			cur_state->alloc);
	}
}

static void emit_w(WORD x)
{
	// no alignment is necessary because all instructions are word length
	expand(2);
	*((WORD*)(cur_state->block->m68k_code + cur_state->out_pos)) = to_BE_w(x);
	cur_state->out_pos += 2;
}

static void emit_l(DWORD x)
{
	expand(4);
	*((DWORD*)(cur_state->block->m68k_code + cur_state->out_pos)) = to_BE_l(x);
	cur_state->out_pos += 4;
}


/* Decoding */
// for changing hl to ix or iy
static int hl2index = 0;

#define get_reg_noindex(n) ((op & (3 << n)) >> n)
#define get_src_reg_noindex() get_reg_noindex(1)
#define get_dst_reg_noindex() get_reg_noindex(4)

static int get_reg(BYTE op, int offset)
{
	int reg = get_reg_noindex(offset);
	if(reg == 2) // hl
		reg += hl2index;
	return reg;
}
#define get_src_reg() get_reg(op, 1)
#define get_dst_reg() get_reg(op, 4)

#define get_bit(n) ((op & (1 << n)) >> n)
#define get_dst_swap() !get_bit(3)
#define get_src_swap() !get_bit(0)

#define get_disp() (hl2index ? read_b() : 0)

/* CCR */
static void save_ccr(void)
{
	if(cur_state->ccr_saved)
		return;
	cur_state->ccr_saved = 1;
	emit_MOVE_from_SR(MODE_Dn, 7);
}

static void mod_ccr(void)
{
	cur_state->ccr_saved = 0;
}

static void restore_ccr(void)
{
	if(!cur_state->ccr_saved)
		return;
	cur_state->ccr_saved = 0;
	emit_MOVE_to_CCR(MODE_Dn, 7);
}

// used by conditional jump/call/return
static void restore_ccr_no_clear(void)
{
	if(!cur_state->ccr_saved)
		return;
	emit_MOVE_to_CCR(MODE_Dn, 7);
}

/* Swapping */
static void set_reg_swap(int reg, BOOL state)
{
	if(reg == 3) // ignore a
		return;
	if(cur_state->reg_swapped[reg] == state)
		return;
	// state is used as the direction to make debugging easier
	// left = swapping on
	// right = swapping off
	// XXX save_ccr not used here b/c unneeded and would produce wasteful
	// code before e.g. INC, DEC which modify flags anyway
	emit_ROd_imm_Dn(state, SIZE_W, 0, reg);
	cur_state->reg_swapped[reg] = state;
}

static void unswap_all(void)
{
	int i;
	for(i = 0; i < 6; i++)
		set_reg_swap(i, 0);
}

static void unswap_all_keepstate(void)
{
	int i;
	for(i = 0; i < 6; i++)
		if(cur_state->reg_swapped[i] && i != 3)
			emit_ROd_imm_Dn(RIGHT, SIZE_W, 0, i);
}

#define set_reg_log_swap(reg,state) (cur_state->reg_swapped[reg] = state)
#define get_reg_swap(reg) (cur_state->reg_swapped[reg])

#ifdef TESTING_ON_PC
#define fix_carry()
#else
// transfer C flag to X before ADC, SBC instructions
#define fix_carry() emit_JSR_longimm(&_fix_carry);
extern void _fix_carry;

asm("
_fix_carry:
bcs	1f
andi	#0xef,%ccr	| clear X
rts
1:
ori	#0x10,%ccr	| set X
rts
");

// emulates ld a,r using rand()
extern void ld_a_r;

asm("
ld_a_r:
movem.l	%d0-%d2/%a0-%a1,-(%sp)
bsr	rand
move.b	%d0,%d3
movem.l	(%sp)+,%d0-%d2/%a0-%a1
tst.b	%d3 			| for flags
rts
");
#endif

// sign-extend a byte to a word
static WORD extend(BYTE b)
{
	if(b & 0x80)
		return (WORD)b | 0xff00;
	else
		return (WORD)b;
}

// to allow branching over unswapping before jump/return
static int get_unswap_len(void)
{
	int i, count;
	for(count = i = 0; i < 6; i++)
		if(get_reg_swap(i) && i != 3)
			count++;
	return count * 2;
}

// needed because of stupid alignment requirement
static void move_16_to_pimm(int reg)
{
	WORD disp = read_w();
	save_ccr();
	if(disp & 1) {
		// 0 then 1, because it's probably just been used as a 16
		set_reg_swap(reg, 0);
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, reg, MODE_pAn_disp16, 1);
		emit_w(disp - 0x8000);
		set_reg_swap(reg, 1);
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, reg, MODE_pAn_disp16, 1);
		emit_w(disp - 0x7fff);
	} else {
		set_reg_swap(reg, 1); // endianness
		emit_MOVE(MOVE_SIZE_W, MODE_Dn, reg, MODE_pAn_disp16, 1);
		emit_w(disp - 0x8000);
	}
}

static void move_16_from_pimm(WORD disp, int reg)
{
	save_ccr();
	if(disp & 1) {
		// 1 then 0, because it'll be probably used as a 16 next
		set_reg_swap(reg, 1);
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp16, 1, MODE_Dn, reg);
		emit_w(disp - 0x7fff);
		set_reg_swap(reg, 0);
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp16, 1, MODE_Dn, reg);
		emit_w(disp - 0x8000);
	} else {
		set_reg_log_swap(reg, 1);
		emit_MOVE(MOVE_SIZE_W, MODE_pAn_disp16, 1, MODE_Dn, reg);
		emit_w(disp - 0x8000);
	}
}

// shadow registers
static WORD shadow_regs[4];

static void exg_with_shadow(int reg)
{
	set_reg_swap(reg, 0);
	emit_MOVE(MOVE_SIZE_W, MODE_OTHER, MODE_OTHER_pimmL, MODE_Dn, 6);
	emit_l((DWORD) &shadow_regs[reg]);
	emit_EXG(Dn_Dn, reg, 6);
	emit_MOVE(MOVE_SIZE_W, MODE_Dn, 6, MODE_OTHER, MODE_OTHER_pimmL);
	emit_l((DWORD) &shadow_regs[reg]);
}

static void translate_op_CB(void);
static BOOL translate_op_ED(void);

BOOL translate_op(void)
{
	BYTE op = read_b();
	int tmp, tmp2;
	void* tmpp;

	switch(op) {
// prefixes
	case 0313: // <prefix> bitwise
		translate_op_CB();
		break;
	case 0355: // <prefix> extra
		return translate_op_ED();
	case 0335: // <prefix> IX
		hl2index = 2;
		tmp = translate_op();
		hl2index = 0;
		return tmp;
	case 0375: // <prefix> IY
		hl2index = 3;
		tmp = translate_op();
		hl2index = 0;
		return tmp;
// ld X,Y
	case 0100: // LD B,B
	case 0111: // LD C,C
	case 0122: // LD D,D
	case 0133: // LD E,E
	case 0144: // LD H,H
	case 0155: // LD L,L
	case 0177: // LD A,A
	case 0102: // LD B,D
	case 0103: // LD B,E
	case 0112: // LD C,D
	case 0113: // LD C,E
	case 0120: // LD D,B
	case 0121: // LD D,C
	case 0130: // LD E,B
	case 0131: // LD E,C
	case 0134: // LD E,H
	case 0135: // LD E,L
	case 0124: // LD D,H
	case 0125: // LD D,L
	case 0114: // LD C,H
	case 0115: // LD C,L
	case 0104: // LD B,H
	case 0105: // LD B,L
	case 0140: // LD H,B
	case 0141: // LD H,C
	case 0142: // LD H,D
	case 0143: // LD H,E
	case 0150: // LD L,B
	case 0151: // LD L,C
	case 0152: // LD L,D
	case 0153: // LD L,E
	case 0107: // LD B,A
	case 0117: // LD C,A
	case 0127: // LD D,A
	case 0137: // LD E,A
	case 0147: // LD H,A
	case 0157: // LD L,A
	case 0170: // LD A,B
	case 0171: // LD A,C
	case 0172: // LD A,D
	case 0173: // LD A,E
	case 0174: // LD A,H
	case 0175: // LD A,L
		save_ccr();
		set_reg_swap(get_src_reg(), get_src_swap());
		set_reg_swap(get_dst_reg(), get_dst_swap());
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, get_src_reg(),
			MODE_Dn, get_dst_reg());
		break;
// ld Xa,Xb
	case 0101: // LD B,C
	case 0110: // LD C,B
	case 0123: // LD D,E
	case 0132: // LD E,D
	case 0145: // LD H,L
	case 0154: // LD L,H
		save_ccr();
		tmp = get_src_reg();
		if(get_reg_swap(tmp) == get_src_swap()) {
			// going low to high
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, tmp, MODE_Dn, 6);
			set_reg_swap(tmp, get_bit(0));
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, 6, MODE_Dn, tmp);
		} else {
			// high to low
			emit_MOVE(MOVE_SIZE_W, MODE_Dn, tmp, MODE_Dn, 6);
			emit_ROd_imm_Dn(RIGHT, SIZE_W, 0, 6);
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, 6, MODE_Dn, tmp);
		}
		break;
// ld X,(HL)
	case 0106: // LD B,(HL)
	case 0116: // LD C,(HL)
	case 0126: // LD D,(HL)
	case 0136: // LD E,(HL)
	case 0156: // LD L,(HL)
	case 0176: // LD A,(HL)
		save_ccr();
		set_reg_swap(2 + hl2index, 0);
		set_reg_swap(get_dst_reg_noindex(), get_dst_swap());
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0,
			MODE_Dn, get_dst_reg_noindex());
		emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG, 2 + hl2index);
		break;
	case 0146: // LD H,(HL)
		save_ccr();
		if(hl2index) {
			set_reg_swap(2 + hl2index, 0);
			emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0,
				MODE_Dn, 2);
			emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG,
				2 + hl2index);
		} else {
			// go through d6
			set_reg_swap(2, 0);
			emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0,
				MODE_Dn, 6);
			emit_pAn_disp8_Rn(0, RDATA, RLONG, 2);
			set_reg_swap(2, 1);
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, 6, MODE_Dn, 2);
		}
		break;
// ld (HL),X
	case 0160: // LD (HL),B
	case 0161: // LD (HL),C
	case 0162: // LD (HL),D
	case 0163: // LD (HL),E
	case 0165: // LD (HL),L
	case 0167: // LD (HL),A
		save_ccr();
		set_reg_swap(2 + hl2index, 0);
		set_reg_swap(get_src_reg_noindex(), get_src_swap());
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, get_src_reg_noindex(),
			MODE_pAn_disp8_Rn, 0);
		emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG, 2 + hl2index);
		break;
	case 0164: // LD (HL),H
		save_ccr();
		if(hl2index) {
			set_reg_swap(2 + hl2index, 0);
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, 2,
				MODE_pAn_disp8_Rn, 0);
			emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG,
				2 + hl2index);
		} else {
			// go through d6
			set_reg_swap(2, 1);
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, 2, MODE_Dn, 6);
			set_reg_swap(2, 0);
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, 6,
				MODE_pAn_disp8_Rn, 0);
			emit_pAn_disp8_Rn(0, RDATA, RLONG, 2);
		}
		break;
// 16-bit lds
	case 0001: // LD BC,nn
	case 0021: // LD DE,nn
	case 0041: // LD HL,nn
		move_16_from_pimm(cur_state->in_pos, get_dst_reg());
		read_w();
		break;
	case 0042: // LD (nn),HL
		move_16_to_pimm(2 + hl2index);
		break;
	case 0052: // LD HL,(nn)
		move_16_from_pimm(read_w(), 2 + hl2index);
		break;
// 8-bit indirect lds
	case 0002: // LD (BC),A
	case 0022: // LD (DE),A
		save_ccr();
		set_reg_swap(get_bit(4), 0);
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, 3,
			MODE_pAn_disp8_Rn, 0);
		emit_pAn_disp8_Rn(0, RDATA, RLONG, get_bit(4));
		break;
	case 0012: // LD A,(BC)
	case 0032: // LD A,(DE)
		save_ccr();
		set_reg_swap(get_bit(4), 0);
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0,
			MODE_Dn, 3);
		emit_pAn_disp8_Rn(0, RDATA, RLONG, get_bit(4));
		break;
	case 0062: // LD (nn),A
		save_ccr();
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, 3, MODE_pAn_disp16, 1);
		emit_w(read_w() - 0x8000u);
		break;
	case 0072: // LD A,(nn)
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp16, 1, MODE_Dn, 3);
		emit_w(read_w() - 0x8000u);
		break;
// 8-bit immediate
	case 0006: // LD B,n
	case 0016: // LD C,n
	case 0026: // LD D,n
	case 0036: // LD E,n
	case 0046: // LD H,n
	case 0056: // LD L,n
	case 0076: // LD A,n
		save_ccr();
		set_reg_swap(get_dst_reg(), get_dst_swap());
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp16, 1,
			MODE_Dn, get_dst_reg());
		emit_w(cur_state->in_pos - 0x8000);
		read_b();
		break;
	case 0066: // LD (HL),n
		save_ccr();
		set_reg_swap(2 + hl2index, 0);
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp16, 1,
			MODE_pAn_disp8_Rn, 0);
		tmp = get_disp();
		emit_w(cur_state->in_pos - 0x8000);
		emit_pAn_disp8_Rn(tmp, RDATA, RLONG, 2 + hl2index);
		read_b();
		break;
// op A,rm8
	// translation table
	static const BYTE get_68k_binop[] = { ADD, 0, SUB, 0, AND, 0, OR, CMP };
	case 0200: // ADD A,B
	case 0201: // ADD A,C
	case 0202: // ADD A,D
	case 0203: // ADD A,E
	case 0204: // ADD A,H
	case 0205: // ADD A,L
	case 0207: // ADD A,A
	case 0220: // SUB B
	case 0221: // SUB C
	case 0222: // SUB D
	case 0223: // SUB E
	case 0224: // SUB H
	case 0225: // SUB L
	case 0227: // SUB A
	case 0240: // AND B
	case 0241: // AND C
	case 0242: // AND D
	case 0243: // AND E
	case 0244: // AND H
	case 0245: // AND L
	case 0247: // AND A
	case 0260: // OR B
	case 0261: // OR C
	case 0262: // OR D
	case 0263: // OR E
	case 0264: // OR H
	case 0265: // OR L
	case 0267: // OR A
	case 0270: // CP B
	case 0271: // CP C
	case 0272: // CP D
	case 0273: // CP E
	case 0274: // CP H
	case 0275: // CP L
	case 0277: // CP A
		mod_ccr();
		set_reg_swap(get_src_reg(), get_src_swap());
		emit_bin_op_ea_Dn(SIZE_B, get_68k_binop[(op >> 3) & 7],
			MODE_Dn, get_src_reg(), 3);
		break;
	case 0206: // ADD A,(HL)
	case 0226: // SUB (HL)
	case 0246: // AND (HL)
	case 0266: // OR (HL)
	case 0276: // CP (HL)
		mod_ccr();
		set_reg_swap(2 + hl2index, 0);
		emit_bin_op_ea_Dn(SIZE_B, get_68k_binop[(op >> 3) & 7],
			MODE_pAn_disp8_Rn, 0, 3);
		emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG, 2 + hl2index);
		break;
	case 0210: // ADC A,B
	case 0211: // ADC A,C
	case 0212: // ADC A,D
	case 0213: // ADC A,E
	case 0214: // ADC A,H
	case 0215: // ADC A,L
	case 0217: // ADC A,A
	case 0230: // SBC A,B
	case 0231: // SBC A,C
	case 0232: // SBC A,D
	case 0233: // SBC A,E
	case 0234: // SBC A,H
	case 0235: // SBC A,L
	case 0237: // SBC A,A
		save_ccr();
		set_reg_swap(get_src_reg(), get_src_swap());
		restore_ccr();
		fix_carry();
		tmp = (op & 0020) ? SUBX : ADDX;
		emit_bin_op_ext_Dn_Dn(SIZE_B, tmp, get_src_reg(), 3);
		mod_ccr();
		break;
	case 0216: // ADC A,(HL)
	case 0236: // SBC A,(HL)
		// through d6
		save_ccr();
		set_reg_swap(2 + hl2index, 0);
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0, MODE_Dn, 6);
		emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG, 2 + hl2index);
		restore_ccr();
		fix_carry();
		tmp = (op & 0020) ? SUBX : ADDX;
		emit_bin_op_ext_Dn_Dn(SIZE_B, tmp, 6, 3);
		mod_ccr();
		break;
	case 0250: // XOR B
	case 0251: // XOR C
	case 0252: // XOR D
	case 0253: // XOR E
	case 0254: // XOR H
	case 0255: // XOR L
	case 0257: // XOR A
		mod_ccr();
		// weird order on the 68K
		set_reg_swap(get_src_reg(), get_src_swap());
		emit_bin_op_ea_Dn(SIZE_B, EOR, MODE_Dn, 3, get_src_reg());
		break;
	case 0256: // XOR (HL)
		mod_ccr();
		// weird order means go through D6
		set_reg_swap(2 + hl2index, 0);
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0, MODE_Dn, 6);
		emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG, 2 + hl2index);
		emit_bin_op_ea_Dn(SIZE_B, EOR, MODE_Dn, 3, 6);
		break;
// op A,imm
	case 0306: // ADD A,n
	case 0326: // SUB n
	case 0346: // AND n
	case 0366: // OR n
	case 0376: // CP n
		mod_ccr();
		emit_bin_op_ea_Dn(SIZE_B, get_68k_binop[(op >> 3) & 7],
			MODE_pAn_disp16, 1, 3);
		emit_w(cur_state->in_pos - 0x8000);
		read_b();
		break;
	case 0316: // ADC A,n
	case 0336: // SBC A,n
		save_ccr();
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp16, 1, MODE_Dn, 6);
		emit_w(cur_state->in_pos - 0x8000);
		read_b();
		restore_ccr();
		fix_carry();
		tmp = (op & 0x38) >> 3; // op
		tmp = (tmp == 1) ? ADDX : SUBX;
		emit_bin_op_ext_Dn_Dn(SIZE_B, tmp, 6, 3);
		break;
	case 0356: // XOR n
		mod_ccr();
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp16, 1, MODE_Dn, 6);
		emit_w(cur_state->in_pos - 0x8000);
		read_b();
		emit_bin_op_ea_Dn(SIZE_B, EOR, MODE_Dn, 3, 6);
		break;
// add16
	case 0011: // ADD HL,BC
	case 0031: // ADD HL,DE
	case 0051: // ADD HL,HL
		mod_ccr(); // XXX: inexact, overflow shouldn't be modified
		tmp = op >> 4;
		if(tmp == 2)
			tmp += hl2index;
		set_reg_swap(2 + hl2index, 0);
		set_reg_swap(tmp, 0);
		emit_bin_op_ea_Dn(SIZE_W, ADD, MODE_Dn, tmp, 2 + hl2index);
		break;
// stack
	case 0305: // PUSH BC
	case 0325: // PUSH DE
	case 0345: // PUSH HL
		save_ccr();
		set_reg_swap(get_dst_reg(), 1); // endianness
		emit_MOVE(MOVE_SIZE_W, MODE_Dn, get_dst_reg(), MODE_pAn_dec, 5);
		break;
	case 0301: // POP BC
	case 0321: // POP DE
	case 0341: // POP HL
		save_ccr();
		set_reg_log_swap(get_dst_reg(), 1);
		emit_MOVE(MOVE_SIZE_W, MODE_pAn_inc, 5, MODE_Dn, get_dst_reg());
		break;
	case 0365: // PUSH AF
		save_ccr();
		emit_ROd_imm_Dn(LEFT, SIZE_W, 0, 6);
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, 3, MODE_Dn, 6);
		emit_MOVE(MOVE_SIZE_W, MODE_Dn, 6, MODE_pAn_dec, 5);
		break;
	case 0361: // POP AF
		emit_MOVE(MOVE_SIZE_W, MODE_pAn_inc, 5, MODE_Dn, 6);
		emit_MOVE(MOVE_SIZE_B, MODE_Dn, 6, MODE_Dn, 3);
		emit_ROd_imm_Dn(RIGHT, SIZE_W, 0, 6);
		cur_state->ccr_saved = 1;
		break;
// incdec16
	case 0003: // INC BC
	case 0023: // INC DE
	case 0043: // INC HL
	case 0013: // DEC BC
	case 0033: // DEC DE
	case 0053: // DEC HL
		save_ccr(); // yes, incdec16 doesn't touch flags
		set_reg_swap(get_dst_reg(), 0);
		emit_bin_op_imm_ea_Q(SIZE_W, (op & 8) >> 3, 1,
			MODE_Dn, get_dst_reg());
		break;
// incdec8
	case 0004: // INC B
	case 0014: // INC C
	case 0024: // INC D
	case 0034: // INC E
	case 0044: // INC H
	case 0054: // INC L
	case 0074: // INC A
	case 0005: // DEC B
	case 0015: // DEC C
	case 0025: // DEC D
	case 0035: // DEC E
	case 0045: // DEC H
	case 0055: // DEC L
	case 0075: // DEC A
		mod_ccr();
		set_reg_swap(get_dst_reg(), get_dst_swap());
		emit_bin_op_imm_ea_Q(SIZE_B, op & 1, 1, MODE_Dn, get_dst_reg());
		break;
	case 0064: // INC (HL)
	case 0065: // DEC (HL)
		mod_ccr();
		set_reg_swap(2 + hl2index, 0);
		emit_bin_op_imm_ea_Q(SIZE_B, op & 1, 1, MODE_pAn_disp8_Rn, 0);
		emit_pAn_disp8_Rn(get_disp(), RDATA, RLONG, 2 + hl2index);
		break;
// rotation
	case 0007: // RLCA
	case 0017: // RRCA
		mod_ccr();
		emit_ROd_imm_Dn((op & 8) ? RIGHT : LEFT, SIZE_B, 1, 3);
		break;
	case 0027: // RLA
	case 0037: // RRA
		restore_ccr();
		emit_ROXd_imm_Dn((op & 8) ? RIGHT : LEFT, SIZE_B, 1, 3);
		mod_ccr();
		break;
// flow
	// translate to inverse condition code, e.g. Z -> NE
	static const BYTE get_68k_inv_cc[] = {
		CC_EQ, CC_NE, CC_CS, CC_CC, 0, /* no parity */ 0, CC_MI, CC_PL
	};
	case 0303: // JP nn
	case 0030: // JR d
		save_ccr();
		unswap_all();
		if(op == 0303)
			tmp = read_w();
		else
			tmp = (WORD) (extend(read_b()) + cur_state->in_pos);
		restore_ccr();
#ifndef TESTING_ON_PC
		tmpp = cur_target->handle_CALL(tmp);
#else
		tmpp = NULL;
#endif
		if(tmpp) {
			emit_JSR_longimm((DWORD) tmpp);
			emit_JMP_longimm(&z80_return);
		} else {
			emit_JMP_longimm((DWORD) translate(tmp));
		}
		return 0;
	case 0302: // JP NZ,nn
	case 0312: // JP Z,nn
	case 0322: // JP NC,nn
	case 0332: // JP C,nn
	case 0362: // JP P,nn
	case 0372: // JP M,nn
	case 0040: // JR NZ,d
	case 0050: // JR Z,d
	case 0060: // JR NC,d
	case 0070: // JR C,d
		if(op & 2) {
			tmp = read_w();
			tmp2 = (op >> 3) & 7;
		} else {
			tmp = (WORD) (extend(read_b()) + cur_state->in_pos);
			tmp2 = (op >> 3) - 4;
		}
#ifndef TESTING_ON_PC
		tmpp = cur_target->handle_CALL(tmp);
#else
		tmpp = NULL;
#endif
		restore_ccr_no_clear();
		emit_Bcc(get_68k_inv_cc[tmp2], get_unswap_len() + 6
			+ (cur_state->ccr_saved ? 2 : 0) + (tmpp ? 4 : 0));
		unswap_all_keepstate();
		restore_ccr();
		if(tmpp) {
			emit_JSR_longimm((DWORD) tmpp);
			emit_JMP_longimm(&z80_return);
		} else {
			emit_JMP_longimm((DWORD) translate(tmp));
		}
		break;
	case 0020: // DJNZ d
		// FIXME: does not handle system calls
		tmp = (WORD) (extend(read_b()) + cur_state->in_pos);
		save_ccr();
		set_reg_swap(0, 1); // decrements b, unfortunately
		emit_bin_op_imm_ea_Q(SIZE_B, SUBQ, 1, MODE_Dn, 0);
		emit_Bcc(CC_EQ, get_unswap_len() + 2 + 6);
		unswap_all_keepstate();
		restore_ccr();
		emit_JMP_longimm((DWORD) translate(tmp));
		break;
	case 0315: // CALL nn
		save_ccr();
		unswap_all();
		tmp = read_w();
#ifndef TESTING_ON_PC
		tmpp = cur_target->handle_CALL(tmp);
#else
		tmpp = NULL;
#endif
		if(tmpp) {
			restore_ccr();
			emit_JSR_longimm(tmpp);
		} else {
			emit_MOVE(MOVE_SIZE_W, MODE_OTHER, MODE_OTHER_imm,
				MODE_pAn_dec, 5);
			emit_w(cur_state->in_pos);
			restore_ccr();
			emit_JMP_longimm(translate(tmp));
			return 0;
		}
		break;
	case 0304: // CALL NZ,nn
	case 0314: // CALL Z,nn
	case 0324: // CALL NC,nn
	case 0334: // CALL C,nn
	case 0364: // CALL P,nn
	case 0374: // CALL M,nn
		tmp2 = read_w();
#ifndef TESTING_ON_PC
		tmpp = cur_target->handle_CALL(tmp2);
#else
		tmpp = NULL;
#endif
		tmp = (op >> 3) & 7;
		restore_ccr_no_clear();
		emit_Bcc(get_68k_inv_cc[tmp], 2 * get_unswap_len() + 6
			+ (cur_state->ccr_saved ? 2 : 0) + (tmpp ? 0 : 4));
		unswap_all_keepstate();
		if(tmpp) {
			restore_ccr();
			emit_JSR_longimm(tmpp);
		} else {
			emit_MOVE(MOVE_SIZE_W, MODE_OTHER, MODE_OTHER_imm,
				MODE_pAn_dec, 5);
			emit_w(cur_state->in_pos);
			restore_ccr();
			emit_JMP_longimm(translate(tmp2));
		}
		unswap_all_keepstate(); // restore swap state
		break;
	case 0311: // RET
		save_ccr();
		unswap_all();
		emit_JMP_longimm(&z80_return);
		return 0;
	case 0300: // RET NZ
	case 0310: // RET Z
	case 0320: // RET NC
	case 0330: // RET C
	case 0360: // RET P
	case 0370: // RET M
		restore_ccr_no_clear();
		emit_Bcc(get_68k_inv_cc[(op & 0030) >> 3],
			get_unswap_len() + (cur_state->ccr_saved ? 2 : 0) + 6);
		unswap_all_keepstate();
		restore_ccr();
		emit_JMP_longimm(&z80_return);
		break;
// miscellanea
	case 0000: // NOP
		emit_NOP();
		break;
	case 0166: // HALT
		save_ccr();
		emit_MOVE(MOVE_SIZE_B, MODE_OTHER, MODE_OTHER_imm,
			MODE_OTHER, MODE_OTHER_pimmL);
		emit_w(0x10); // wake on auto-int 5
		emit_l(0x600005); // port to put CPU to sleep
		break;
	case 0057: // CPL
		save_ccr();
		emit_NOT(SIZE_B, MODE_Dn, 3);
		break;
	case 0067: // SCF
		emit_ORI_to_CCR(1);
		break;
	case 0077: // CCF
		emit_EORI_to_CCR(1);
		break;
	case 0353: // EX DE,HL
		// HL is still used even with IX/IY prefix
		set_reg_swap(1, get_reg_swap(2));
		emit_EXG(Dn_Dn, 1, 2);
		break;
	case 0343: // EX (SP),HL
		// the assumption is safe that SP is word-aligned
		save_ccr();
		set_reg_swap(2 + hl2index, 1); // endianness
		emit_MOVE(MOVE_SIZE_W, MODE_pAn, 5, MODE_Dn, 6);
		emit_EXG(Dn_Dn, 2 + hl2index, 6);
		emit_MOVE(MOVE_SIZE_W, MODE_Dn, 6, MODE_pAn, 5);
		break;
	case 0010: // EX AF,AF'
		save_ccr();
		exg_with_shadow(3);
		break;
	case 0331: // EXX
		save_ccr();
		exg_with_shadow(0);
		exg_with_shadow(1);
		exg_with_shadow(2);
		break;
	case 0363: // DI
		// some programs disable interrupts to use the shadow regs
		// or get speed... safe to make a no-op
		break;
	case 0373: // EI
		// some (read: all good) programs use this before HALT just in
		// case... safe to make a no-op
		break;
// HW-dependent
	case 0323: // OUT (n),A
		save_ccr();
		emit_JSR_longimm(cur_target->handle_OUT(read_b()));
		break;
	case 0333: // IN A,(n)
		save_ccr();
		emit_JSR_longimm(cur_target->handle_IN(read_b()));
		break;
	case 0307: // RST 00
	case 0317: // RST 08
	case 0327: // RST 10
	case 0337: // RST 18
	case 0347: // RST 20
	case 0357: // RST 28
	case 0367: // RST 30
	case 0377: // RST 38
		save_ccr();
		unswap_all();
		tmpp = cur_target->handle_RST(op);
		restore_ccr();
		emit_JSR_longimm(tmpp);
		break;
// unsupported
	// parity - 68K doesn't have this
	// case 0340: // RET PO
	// case 0350: // RET PE
	// case 0344: // CALL PO,nn
	// case 0354: // CALL PE,nn
	// case 0342: // JP PO,nn
	// case 0352: // JP PE,nn
	// SP
	// case 0061: // LD SP,nn
	// case 0371: // LD SP,HL
	// case 0063: // INC SP
	// case 0073: // DEC SP
	// case 0071: // ADD HL,SP
	// other
	// case 0047: // DAA
	// case 0351: // JP (HL)
		// despite the mnemonic, this is HL->PC, not (HL)->PC
		// and so there is no displacement with IX/IY
		// (but IX/IY are still used if prefixed)
	default:
		report_error("unsupported op %.2X", op);
		// does not return
	}

	return 1;
}

/* Rotates, shifts, and single-bit operations.
 * You are not expected to understand this.
 */
static void translate_op_CB(void)
{
	WORD disp = get_disp();
	BYTE op = read_b();

	// FIXME: On the Z80, IX/IY prefix + CB instruction not on (HL) means
	// "perform instruction on index+n and store result to register".
	// This is not supported.
	if(hl2index > 0 && (op & 7) != 6)
		report_error("unsupported CB op");

	// rotate or shift
	if(op < 0x40) {
		static const WORD rotate_shift_op[] = { ROd, ROXd, ASd, LSd };
		WORD op_68k = rotate_shift_op[op >> 4];
		if((op & 7) != 6) {
			// register
			save_ccr();
			set_reg_swap(get_src_reg_noindex(), get_src_swap());
			restore_ccr();
			emit_RoSh_imm_Dn(op_68k, (op & 8) ? RIGHT : LEFT,
				SIZE_B,	1, get_src_reg_noindex());
			if(op >> 3 == 6) // SLL
				emit_bin_op_imm_ea_Q(SIZE_B, ADDQ, 1,
					MODE_Dn, get_src_reg_noindex());
			mod_ccr();
		} else {
			// (HL)
			save_ccr();
			set_reg_swap(2 + hl2index, 0);
			emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0,
				MODE_Dn, 6);
			emit_pAn_disp8_Rn(disp, RDATA, RLONG, 2 + hl2index);
			restore_ccr();
			emit_RoSh_imm_Dn(op_68k, (op & 8) ? RIGHT : LEFT,
				SIZE_B, 1, 6);
			if(op >> 3 == 6) // SLL
				emit_bin_op_imm_ea_Q(SIZE_B, ADDQ, 1,
					MODE_Dn, 6);
			mod_ccr();
			save_ccr();
			emit_MOVE(MOVE_SIZE_B, MODE_Dn, 6,
				MODE_pAn_disp8_Rn, 0);
			emit_pAn_disp8_Rn(disp, RDATA, RLONG, 2 + hl2index);
		}
	} else {
		// single-bit
		static const WORD bit_op[] = { 0, BTST, BCLR, BSET };
		WORD op_68k = bit_op[op >> 6];
		int bitnr = (op >> 3) & 7;
		if(op >> 6 == 1)
			mod_ccr();
		else
			save_ccr();
		if((op & 7) != 6) {
			// register
			set_reg_swap(get_src_reg_noindex(), get_src_swap());
			emit_bitop_imm(op_68k, bitnr,
				MODE_Dn, get_src_reg_noindex());
		} else {
			// (HL)
			set_reg_swap(2 + hl2index, 0);
			emit_bitop_imm(op_68k, bitnr, MODE_pAn_disp8_Rn, 0);
			emit_pAn_disp8_Rn(disp, RDATA, RLONG, 2 + hl2index);
		}
	}
}

/* Sundry extra operations. */
static BOOL translate_op_ED(void)
{
	BYTE op = read_b();

	switch(op) {
#ifdef TESTING_ON_PC
	case 0: // terminator in testing
		emit_w(0xDEAD);
		return 0;
#endif
// ld16
	case 0103: // LD (nnnn),BC
	case 0123: // LD (nnnn),DE
		move_16_to_pimm(get_bit(4));
		break;
	case 0113: // LD BC,(nnnn)
	case 0133: // LD DE,(nnnn)
		move_16_from_pimm(read_w(), get_bit(4));
		break;
// io
	// case 0100: // IN B,(C)
	// case 0101: // OUT (C),B
	// case 0110: // IN C,(C)
	// case 0111: // OUT (C),C
	// case 0120: // IN D,(C)
	// case 0121: // OUT (C),D
	// case 0130: // IN E,(C)
	// case 0131: // OUT (C),E
	// case 0140: // IN H,(C)
	// case 0141: // OUT (C),H
	// case 0150: // IN L,(C)
	// case 0151: // OUT (C),L
	// case 0170: // IN A,(C)
	// case 0171: // OUT (C),A
// op16
	case 0102: // SBC HL,BC
	case 0112: // ADC HL,BC
	case 0122: // SBC HL,DE
	case 0132: // ADC HL,DE
	case 0142: // SBC HL,HL
	case 0152: // ADC HL,HL
		save_ccr();
		set_reg_swap(2, 0);
		set_reg_swap(get_dst_reg(), 0);
		restore_ccr();
		emit_ORI_to_CCR(4); // SUBX will never set Z
		emit_bin_op_ext_Dn_Dn(SIZE_W, (op & 8) ? ADDX : SUBX,
			get_dst_reg(), 2);
		mod_ccr();
		break;
// special regs
	// case 0107: // LD I,A
	// case 0127: // LD A,I
	case 0117: // LD R,A
		// some programs use this to get better entropy...
		// not needed with actual RNG!
		break;
#ifndef TESTING_ON_PC
	case 0137: // LD A,R
		emit_JSR_longimm(&ld_a_r);
		break;
#endif
// interrupt
	// case 0106: // IM 0
	// case 0126: // IM 1
	// case 0136: // IM 2
	// case 0115: // RETI
	// case 0105: // RETN
// rotate/shift
	// case 0147: // RRD
	// case 0157: // RLD
		// rotate [A,(HL)] right/left 4 bits
// SIMD
	case 0240: // LDI
	case 0250: // LDD
	case 0260: // LDIR
	case 0270: // LDDR
		save_ccr();
		set_reg_swap(0, 0);
		set_reg_swap(1, 0);
		set_reg_swap(2, 0);
		if(op & 0020) {
			emit_BRA(10); // branch to DBRA
		}
		emit_MOVE(MOVE_SIZE_B, MODE_pAn_disp8_Rn, 0,
			MODE_pAn_disp8_Rn, 0);
		emit_pAn_disp8_Rn(0, RDATA, RLONG, 2);
		emit_pAn_disp8_Rn(0, RDATA, RLONG, 1);
		emit_bin_op_imm_ea_Q(SIZE_W, (op & 0010) ? SUBQ : ADDQ, 1,
			MODE_Dn, 1);
		emit_bin_op_imm_ea_Q(SIZE_W, (op & 0010) ? SUBQ : ADDQ, 1,
			MODE_Dn, 2);
		if(op & 0020) {
			emit_DBRA(0, NEG_WORD(12));
			// BC=-1 at exit
			emit_bin_op_imm_ea_Q(SIZE_W, ADDQ, 1, MODE_Dn, 0);
		}
		break;
	// case 0241: // CPI
	// case 0251: // CPD
	// case 0261: // CPIR
	// case 0271: // CPDR
	// case 0242: // INI
	// case 0252: // IND
	// case 0262: // INIR
	// case 0272: // INDR
	// case 0243: // OUTI
	// case 0253: // OUTD
	// case 0263: // OTIR
	// case 0273: // OTDR
// sp
	// case 0172: // ADC HL,SP
	// case 0162: // SBC HL,SP
	// case 0163: // LD (nnnn),SP
	// case 0173: // LD SP,(nnnn)
// misc
	case 0104: // NEG
		mod_ccr();
		emit_NEG(SIZE_B, MODE_Dn, 3);
		break;
	default:
		report_error("unsupported op ED %.2X", op);
		// does not return
	}

	return 1;
}
