/*
 * TI-83+ target.
 *
 * Copyright (C) 2006 Nick Tarleton.
 * Licensed under the GNU General Public License; see COPYING.txt.
 */

#include <string.h>
#include <alloc.h>

#include "misc.h"
#include "target.h"
#include "xlate.h"
#include "glue.h"

#include "ti83plus/funcs.h"
#include "ti83plus/ti83plus.inc.h"
#include "ti83plus/ion.inc.h"
#include "ti83plus/mirage.inc.h"

static M68K_CODE translated;
static WORD size;
static BYTE* start_of_stored_prog;

static void load_83p(BYTE* fdata)
{
	BYTE* reading_fdata;

	size = *(WORD *)fdata;
	fdata += 2;

	if(*(WORD*)fdata != 0xBB6D) // is ASM?
		report_error("Not an ASM program. Enki cannot run BASIC programs.");
	fdata += 2;
	start_of_stored_prog = reading_fdata = fdata;
	
	// get type in detail
	if(*reading_fdata == 0311) { // RET - indicates shell-dependent prog
		reading_fdata++;
		if(*reading_fdata == 0060) { // JR NC - indicates Ion
			// skip description so it's not 'translated'
			// (it would be because the jump is conditional)
			reading_fdata++;
			reading_fdata += *reading_fdata + 1;
		} else if(*reading_fdata == 1) { // indicates Mirage normal prog
			reading_fdata += 33; // skip icon
			while(*(reading_fdata++)) ; // skip description
		} else { // unknown format
			report_error("Unknown ASM program format. May be a data file (e.g. a level), a Mirage external interface, a Mirage program using unsupported features, or a program for an unsupported shell.");
		}
	} else if(reading_fdata[0] == 0257 && reading_fdata[1] == 0060) {
		// XOR A, JR NC - Ion
		reading_fdata += 2;
		reading_fdata += *reading_fdata;
	} else if(reading_fdata[0] == 'c' && reading_fdata[1] == 'R'
		&& reading_fdata[2] == 'u') {
		// CrunchyOS - really should use a RET!!!
		report_error("CrunchyOS program, not supported");
	} // else a normal ASM program (we hope!)
	
	// set up memory space
	z80_base = malloc_remember(0x8000) - 0x8000;
	memset(z80_base + 0x8000, 0, 0x8000); // clear everything
	memcpy(z80_base + TI83P_progstart, start_of_stored_prog, size);
		// load program
	// set up flags
	// FIXME many, many flags are not supported
	z80_base[TI83P_flags + TI83P_appflags] = 1 << TI83P_appAutoScroll;

	// start translating at start of actual code
	translated = translate(TI83P_progstart + (reading_fdata - fdata));
}

static const WIN_RECT viewport = {30, 16, 129, 83};

static void run_83p(void)
{
	Z80_REGS regs;
	regs.words.iy = TI83P_flags;
	regs.words.sp = 0xFFFE; // don't know the real value but this should do
	key_was_pressed_last_time = 0;
	DrawClipRect(&viewport,	&(SCR_RECT){{0, 0, 159, 99}},
		A_NORMAL | B_DOUBLE);
	// cheaply clear screen
	asm("bsr clear_viewport" : : : "a2", "d6", "cc");
	call_gen(translated, &regs);
	
	// write back
	memcpy(start_of_stored_prog, z80_base + TI83P_progstart, size);
}

static void* RST_83p(BYTE op)
{
	WORD bcall;

	if(op != 0357) { // RST 28h = bcall
		switch(op) {
			// case 0307: // 00h = reset
			case 0317: // 08h = op1toop2
				return &TI83P_func_op1toop2;
			// case 0327: // 10h = findsym
			// case 0337: // 18h = pushrealo1
			case 0347: // 20h = mov9toop1
				return &TI83P_func_mov9toop1;
			// case 0367: // 30h = fpadd
			// case 0377: // 38h = ISR
			default:
				report_error("unsupported RST %.2X", op);
		}
	}

	// bcall
	bcall = read_w();
	switch(bcall) {
		case TI83P_grbufclr:
			return &TI83P_func_grbufclr;
		case TI83P_clrlcdfull:
		case TI83P_clrscrnfull: // txtshad is not emulated
			return &TI83P_func_clrlcdfull;
		case TI83P_grbufcpy:
			return &TI83P_func_grbufcpy;
		case TI83P_darkline:
			return &TI83P_func_darkline;
		case TI83P_runindicon:
		case TI83P_runindicoff:
			// no-op - there is no run indicator
			return &no_op;
		case TI83P_homeup:
			return &TI83P_func_homeup;
		case TI83P_newline:
			return &TI83P_func_newline;
		case TI83P_putmap:
			return &TI83P_func_putmap;
		case TI83P_putc:
			return &TI83P_func_putc;
		case TI83P_puts:
			return &TI83P_func_puts;
		case TI83P_disphl:
			return &TI83P_func_disphl;
		case TI83P_putps:
			return &TI83P_func_putps;
		case TI83P_vputmap:
			return &TI83P_func_vputmap;
		case TI83P_vputs:
			return &TI83P_func_vputs;
		case TI83P_vputsn:
			return &TI83P_func_vputsn;
		case TI83P_getcsc:
		case TI83P_getk: // XXX not exactly the same (but close)
			return &TI83P_func_getcsc;
		case TI83P_enoughmem:
			return &TI83P_func_enoughmem;
		case TI83P_cphlde:
			return &TI83P_func_cphlde;
		case TI83P_divhlby10:
			return &TI83P_func_divhlby10;
		case TI83P_divhlbya:
			return &TI83P_func_divhlbya;
		case TI83P_mov9toop1:
			return &TI83P_func_mov9toop1;
		case TI83P_op1toop2:
			return &TI83P_func_op1toop2;
		case TI83P_op2toop1:
			return &TI83P_func_op2toop1;
		case TI83P_setxxop1:
			return &TI83P_func_setxxop1;
		case TI83P_setxxxxop2:
			return &TI83P_func_setxxxxop2;
		case TI83P_dispop1a:
			return &TI83P_func_dispop1a;
		default:
			report_error("unsupported BCALL %.2X", bcall);
	}
}

static void* CALL_83p(WORD addr)
{
	if(addr >= TI83P_progstart)
		return NULL;

	switch(addr) {
		case TI83P_ionVersion:
		case TI83P_iversion:
			return &TI83P_func_ionVersion;
		case TI83P_ionRandom:
		case TI83P_irandom:
			return &TI83P_func_ionRandom;
		case TI83P_ionGetPixel:
		case TI83P_igetpix:
			return &TI83P_func_ionGetPixel;
		case TI83P_ionFastCopy:
		case TI83P_ifastcopy:
			return &TI83P_func_grbufcpy;
		case TI83P_ionPutSprite:
		case TI83P_isprite:
			return &TI83P_func_ionPutSprite;
		case TI83P_MIRAGE_cphlde:
			return &TI83P_func_cphlde;
		default:
			report_error("unsupported CALL %.2X", addr);
	}
}

static void* IN_83p(BYTE port)
{
	switch(port) {
		case 1:
			return &TI83P_in_kbd;
		default:
			report_error("unsupported IN %.2X", port);
	}
}

static void* OUT_83p(BYTE port)
{
	switch(port) {
		case 1:
			return &TI83P_out_kbd;
		case 3: // used to turn off calc, ignore
			return &no_op;
		default:
			report_error("unsupported OUT %.2X", port);
	}
}

TARGET target_ti83plus = {
	load_83p,
	run_83p,
	RST_83p,
	CALL_83p,
	IN_83p,
	OUT_83p,
};
