/*
 * Translation core.
 * note: all the weird stuff with TESTING_ON_PC is necessary because this file
 * can be compiled either for the calculator or for the computer (in order to
 * test translation in a more convenient environment)
 *
 * Copyright (C) 2006 Nick Tarleton.
 * Licensed under the GNU General Public License; see COPYING.txt.
 */

#ifndef TESTING_ON_PC
#	include <tigcclib.h>
#else
#	include <stdlib.h>
#	include <string.h>
#endif

#include "xlate.h"
#include "misc.h"
#include "ops.h"
#include "68k_gen.h"

/* Tunable memory-allocation parameters */
#define INITIAL_CODE_ALLOC 32 // space to allocate initially for translated code
#define SHRINK_CODE_WHEN_DONE // if defined, code blocks will be realloced to minimum size when complete
#define BLOCK_ALLOC_SIZE 10 // number of BLOCKs to allocate at once

// this is actually the top of a stack
XLATE_STATE* cur_state;

/* Access */

BYTE read_b(void)
{
	return z80_base[cur_state->in_pos++];
}

WORD read_w(void)
{
	// *()&(*)&)(! ALIGNMENT REQUIREMENT!
	WORD rv = read_b();
	rv |= read_b() << 8;
	return rv;
}

/* Block allocator
 * This gets around the efficiency problem malloc() has with small objects.
 * It allocates enough space to hold BLOCK_ALLOC_SIZE BLOCKs at a time, then
 * parcels that out.
 */
static BLOCK* free_blocks;
static int rem_free_blocks;

// Allocate one block.
static BLOCK* alloc_block(void)
{
	if(rem_free_blocks == 0) {
		rem_free_blocks = BLOCK_ALLOC_SIZE;
		free_blocks = malloc_remember(rem_free_blocks * sizeof(BLOCK));
	}
	rem_free_blocks--;
	return free_blocks++;
}

/* Translated cache
 * Simple unordered linked list.
 */
static BLOCK* translated_head;

/* Find a block in the cache, or return NULL if none. */
static M68K_CODE scan_block(Z80_LOC loc)
{
	BLOCK* block;
	for(block = translated_head; block != NULL; block = block->next)
		if(block->z80_start == loc)
			return block->m68k_code;
	return 0;
}

/* Translation */
M68K_CODE translate(Z80_LOC loc)
{
	XLATE_STATE local_state;
	XLATE_STATE* state;
	BLOCK* block;

	// search
	M68K_CODE search = scan_block(loc);
	if(search)
		return search;

	// search blocks currently being translated
	for(state = cur_state; state != NULL; state = state->next)
		if(state->block->z80_start == loc)
			// we don't know where the block will end up once
			// completely translated, so use a trampoline...
			return (M68K_CODE) &(state->block->trampoline)
				+ add_ghost_space;
			// ... which will be patched later with the final addr.
			// Elegant, no?
	
	block = alloc_block();
	local_state.block = block;

	block->z80_start = loc;
	local_state.in_pos = loc;
	local_state.out_pos = 0;
	block->m68k_code = malloc_remember(INITIAL_CODE_ALLOC);
	local_state.alloc = INITIAL_CODE_ALLOC;
	memset(&local_state.reg_swapped, 0, sizeof(local_state.reg_swapped));
	local_state.ccr_saved = 0;

	local_state.next = cur_state;
	cur_state = &local_state; // push

	while(translate_op()) {
#ifndef TESTING_ON_PC
		if(!peekIO_bit(0x60001a, 1)) // if ON pressed
			ER_throw(ER_BREAK);
#endif
	}

#ifdef SHRINK_CODE_WHEN_DONE
	block->m68k_code = realloc_remember(block->m68k_code,
		local_state.out_pos);
#endif
	block->m68k_code += add_ghost_space;
	// offset allows execution in ghost space on HW2

	block->next = translated_head;
	translated_head = block;

	// create trampoline
	block->trampoline.jmp_instr = to_BE_w(0x4EF9); // JMP imm32
	block->trampoline.target = to_BE_l((DWORD) block->m68k_code);

	cur_state = local_state.next; // pop
	return block->m68k_code;
}

void xlate_zero_statics(void)
{
	rem_free_blocks = 0;
	translated_head = NULL;
	cur_state = NULL;
}
